$:.unshift File.join(File.dirname(__FILE__), 'lib')

require 'sinatra/base'
require 'emcee'
require 'json'
require 'digest/md5'
require 'resource_management'
require 'parameter_validation'

class LiveManagement < Sinatra::Base


  @@plain_string_rule = ParameterValidation::Rule.new(String)
  @@plain_hash_rule = ParameterValidation::Rule.new(Hash)
  @@plain_array_rule = ParameterValidation::Rule.new(Array)
  @@null_rules = {}

  @@mc_run_rules = {
    'nodeset_id'  => @@plain_string_rule,
    'agent_name'  => @@plain_string_rule,
    'action_name' => @@plain_string_rule,
    'input'       => ParameterValidation::Rule.new(Hash, true, ParameterValidation.method(:keys_to_symbols))
  }

  @@get_nodes_rules = {
    'facts'      => ParameterValidation::Rule.new(Hash, true, ParameterValidation.method(:fact_formatter)),
    'classes'    => ParameterValidation::Rule.new(String, true, ParameterValidation.method(:case_insensitive)),
    'identities' => ParameterValidation::Rule.new(String, true, ResourceManagement.method(:string_to_nodes_regex))
  }

  @@resources_by_type_rules = {
    'nodeset_id' => @@plain_string_rule,
    'type'       => @@plain_string_rule
  }

  @@resources_by_type_and_title_rules = {
    'nodeset_id' => @@plain_string_rule,
    'type'       => @@plain_string_rule,
    'title'      => @@plain_string_rule
  }

  @@clone_rules = {
    'type'         => @@plain_string_rule,
    'title'        => @@plain_string_rule,
    'parameters'   => @@plain_hash_rule,
    'nodeset_id'   => @@plain_string_rule,
    'origin_nodes' => @@plain_array_rule
  }

  @@post_nodeset_rules = {
    'nodes' => ParameterValidation::Rule.new(Array, true)
  }

  @@get_nodeset_rules = {
    'key' => @@plain_string_rule
  }

  # the app defaults to development by default
  # raise errors helps a lot with knowing why tests failed
  # to run in production set RACK_ENV=production before starting the web server
  configure :development, :test do
    enable :raise_errors
  end

  def cache_strategy
    :memcache # valid choices are :memcache, :file, :none, or nil (=> use default of memcache w/ file fallback)
  end

  def emcee(args = {})
    Emcee.new({ :cache => cache_strategy }.merge(args))
  end

  # XXX : Shouldn't use the Rails logger any more.
  def logger
    @logger ||= if defined?(Rails) && Rails.logger
      Rails.logger
    else
      require 'logger'
      Logger.new("log/#{settings.environment}.log")
    end
  end

  set :root, File.dirname(__FILE__)
  enable :logging

  get '/' do
    erb :index
  end

  get '/lm/mc_agents' do
    cache_control :no_cache
    content_type :json
    errors = ParameterValidation.validate(params, @@null_rules)
    return [400, errors.join('\n')] unless errors.empty?
    # Remove create action from puppetral agent.
    agents = JSON.parse(emcee().agents.to_json)
    agents.each do |agent|
      if agent['name'] == 'puppetral'
        agent['actions'].reject! { |action| action['name'] == 'create' }
      end
    end
    agents.to_json
  end

  post '/lm/mc_run' do
    cache_control :no_cache
    content_type :json
    begin
      nodes = nodeset_error_check(params, @@mc_run_rules)
    rescue Exception => e
      return [400, e.message]
    end
    mc = emcee(:nodeset => nodes)
    agent = mc.agents_by_name(params[:agent_name]).first
    return [400, "MCollective agent not found."] unless agent
    action = agent.action_by_name(params[:action_name])

    # Run the action and return the results.
    input = params[:input]
    if input
      results = action.run(input)
    else
      results = action.run
    end

    grouped_results = Hash.new { |hash, key| hash[key] = Array.new }
    results.each { |r| grouped_results[r.contents] << r.node_name }

    # needs to be an array else hashes as hash keys get stringified
    # ex. { {:key1 => 'a'} => 'first', {:key1 => 'b'} => 'second'}.to_json = "{\"key1a\":\"first\",\"key1b\":\"second\"}"
    # and we sort it so that the most common results are first
    return grouped_results.sort_by {|k,v| v.size}.reverse.to_json
  end

  get '/lm/nodes' do
    cache_control :no_cache
    content_type :json
    errors = ParameterValidation.validate(params, @@get_nodes_rules)
    return [400, errors.join('\n')] unless errors.empty?
    params.delete('facts') unless params['facts']
    nodes = emcee(:filter => params).filtered_nodes.sort
    nodeset_id = ResourceManagement.generate_nodeset_id(nodes)
    {'id' => nodeset_id, 'nodes' => nodes.map {|n| { 'name' => n } } }.to_json
  end

  get '/lm/resources/:type' do
    cache_control :no_cache
    content_type :json
    begin
      query_nodes = nodeset_error_check(params, @@resources_by_type_rules)
    rescue Exception => e
      return [400, e.message]
    end
    if query_nodes.empty?
      resources, resource_count, variation_count = [], 0, 0
    else
      resources, resource_count, variation_count = ResourceManagement.fetch(query_nodes, params[:type])
    end
    { :total      => resource_count,
      :variations => variation_count,
      :resources  => resources
    }.to_json
  end

  get '/lm/resources/:type/:title' do
    cache_control :no_cache
    content_type :json
    begin
      query_nodes = nodeset_error_check(params, @@resources_by_type_and_title_rules)
    rescue Exception => e
      return [400, e.message]
    end
    title = params[:title]
    if query_nodes.empty?
      resources, resource_count, variation_count = [{ :variations => [],
                                                      :title      => title,
                                                      :nodeCount  => 0,
                                                      :properties => []
                                                    }], 0, 0
    else
      resources, resource_count, variation_count = ResourceManagement.fetch(query_nodes, params[:type], title)
    end
    result = resources[0]
    return result ? result.to_json : [400, "No matching resource was found"]
  end

  post '/lm/clone' do
    cache_control :no_cache
    content_type :json
    begin
      target_nodes = nodeset_error_check(params, @@clone_rules)
    rescue Exception => e
      return [400, e.message]
    end
    origin_nodes = params['origin_nodes']
    resource = {
      'type'       => params['type'],
      'title'      => params['title'],
      'parameters' => params['parameters']
    }
    original_resource = resource.dup
    ResourceManagement.enforce_parameter_whitelist!(resource)

    # Don't copy over the resource on the origin nodeset - this is
    # guaranteed noop.
    target_nodes -= origin_nodes

    # Perform the clone operation and format and return the results.
    results = ResourceManagement.clone(target_nodes, resource)
    return [500, "Something wildly unexpected happened. Please file a bug at http://projects.puppetlabs.com/projects/puppet-enterprise/issues"] if results[:successes].length > 1
    clone_success = results[:successes].first
    properties = clone_success ? clone_success['properties'].map{|p| p.to_s} : []
    properties |= original_resource['parameters'].keys.map{|p| p.to_s}
    properties.sort!
    results[:properties] = properties
    results[:original] = original_resource
    results[:successes] = clone_success ? clone_success['variations'] : []
    results[:counts] = {
      :successes => clone_success ? clone_success['nodeCount'] : 0,
      :failures  => results[:failures].length,
      :errors    => results[:errors].length
    }

    results.to_json
  end

  post '/lm/clone/preview' do
    cache_control :no_cache
    content_type :json
    begin
      target_nodes = nodeset_error_check(params, @@clone_rules)
    rescue Exception => e
      return [400, e.message]
    end
    origin_nodes = params['origin_nodes']
    resource = {
      'type'       => params['type'],
      'title'      => params['title'],
      'parameters' => params['parameters']
    }
    ResourceManagement.enforce_parameter_whitelist!(resource)

    # Don't copy over the resource on the origin nodeset - this is
    # guaranteed noop.
    target_nodes -= origin_nodes
    results = { :creations => [], :failures => [], :overwrites => [],
                :properties => [], :counts => { :creations => 0,
                                                :failures => 0,
                                                :overwrites => 0 }
    }
    return results.to_json if target_nodes.empty?

    mc = emcee(:nodeset => target_nodes)
    mc_resources = mc.resource(params[:type], params[:title], true)
    nodes_with_resource = mc_resources.map {|res| res['node_name']}
    ResourceManagement.frontend_encode_resources!(target_nodes, mc_resources)

    # Here is where we would check nodes in the target nodeset for potential failures
    results[:creations] = (target_nodes - nodes_with_resource).collect do |name|
      results[:counts][:creations] += 1
      { :nodeName => name }
    end
    overwrites = []
    mc_resources.each do |res|
      ResourceManagement.enforce_parameter_whitelist!(res)
      if res['managed'] == true
        results[:failures] << { :nodeName => res["node_name"],
                                :reason   => "Resource is managed by Puppet." }
        results[:counts][:failures] += 1
      else
        overwrites << res
        results[:properties] |= res["parameters"].keys.map{|p| p.to_s}
        results[:counts][:overwrites] += 1
      end
    end
    results[:overwrites] = overwrites.empty? ? [] : ResourceManagement.resources_to_variations(overwrites)[0][0]['variations']
    results[:properties] |= params['parameters'].keys.map{|p| p.to_s}
    results[:properties].uniq!
    results[:properties].sort!
    results.to_json
  end

  get '/lm/resource_types' do
    cache_control :no_cache
    content_type :json
    errors = ParameterValidation.validate(params, @@null_rules)
    return [400, errors.join('\n')] unless errors.empty?
    ['user','package','host','group'].collect do |type|
      {:name => type}
    end.to_json
  end

  post '/lm/nodeset' do
    errors = ParameterValidation.validate(params, @@post_nodeset_rules)
    return [400, errors.join('\n')] unless errors.empty?
    ResourceManagement.generate_nodeset_id(params[:nodes] || [])
  end

  get '/lm/nodeset' do
    errors = ParameterValidation.validate(params, @@get_nodeset_rules)
    return [400, errors.join('\n')] unless errors.empty?
    nodeset = ResourceManagement.retrieve_nodeset(params[:key])
    if nodeset
      return nodeset.sort.to_json
    else
      [400, "No such nodeset is cached"]
    end
  end

  def nodeset_error_check(params, rules)
    errors = ParameterValidation.validate(params, rules)
    raise errors.join('\n') unless errors.empty?
    nodeset_id = params['nodeset_id']
    nodes = ResourceManagement.retrieve_nodeset(nodeset_id)
    raise ResourceManagement::NO_NODESET_ERROR unless nodes
    return nodes
  end

end
